/*--------------------------------------------------------------------------
 *    
 *                         -- C N R S -- 
 *         Laboratoire d'Automatique et d'Analyse des Systemes 
 *                 7 Avenue du colonel Roche   
 *                     31 077 Toulouse Cedex
 *
 *  Auteur           : Sara Fleury
 *  Groupe           : Robotique et Intelligence Artificielle
 *  Nom du fichier   : piloCmdTaskFunc.c
 *  Date de creation : Juin 1992
 * 
 *  Description      : Calcul de l'odometrie "probabiliste"
 *                     Trs inspir de odoProba.c du module loco de H2
 *
 *  Copyright (C) 1992-2001 LAAS/CNRS.
 *----------------------------------------------------------------------*/
#include <stdio.h>
#include <h2mathLib.h>
#include "scoutStruct.h"

static void Calc_var_whee(double ds, double dth, double lax, 
			  double coeff, double *vdep);

static void calc_nouv_pos(SCOUT_POS_STR *posId, 
			  double dS, double dTheta,
			  double LaX,
			  double *jxr, double *jdep);


#define Coef 0.002			  /* pourcentage d'erreur sur le
					   * deplacement de chacune des
					   * roues; la valeur indiquee est
					   * plus qu'aproximative !! A mon
					   * avis il faut faire des essais
					   * ... (c'est beau la science !!!) */

static int dms_plus(const double *m1, const double *m2, double *m_res,
		int l, int c);
static int dmfsf_jmult2(const double *m1, const double *m2, double *m_res,
		int l1, int c1, int l2, int c2);


/* jacobiennes (J_h_Xr,J_h_dR) */
static double J_h_Xr[9], J_h_dR[6];

/* tableaux auxilliaires*/
static double aux[9];
static double V_dep[3];

/**
 ** fonction de calcul de l'odometrie
 **
 ** entree: posId - nouvelle position + incertitude 
 **                 deplacements mesures
 **
 ** sortie: posId - incertitude
 **/

void odoProba(SCOUT_POS_STR *posId,  	/* ident poster Odometrie */
	      double dS, double dTh,
	      double lax, 	/* entre-axe */
	      double coeff,
	      SCOUT_ODO_VAR_STR *var)
{

    /*
     * calcul de la nouvelle position et des jacobiennes en fonction des
     * deplacements des roues.
     */
    calc_nouv_pos(posId, dS, dTh, lax, J_h_Xr, J_h_dR);

    /*
     * variance sur le deplacement
     */
    Calc_var_whee(dS, dTh, lax, coeff, V_dep);

    /*
     * calcul de la variance sur Xrob
     */
    dmfsf_jmult2(J_h_dR, V_dep, aux, 3, 2, 2, 2);
    dmfsf_jmult2(J_h_Xr, var->var, var->var, 3, 3, 3, 3);
    dms_plus(aux, var->var, var->var, 3, 3);

}

/*----------------------------------------------------------------------*/

/**
 ** fonction "h" de l'odometrie + jacobiennes
 **/

static void calc_nouv_pos(SCOUT_POS_STR *posId, 
			  double dS, double dTheta,
			  double LaX,
			  double *jxr, double *jdep)
{
  double s, c;
  double thetaMoy;

  thetaMoy = posId->theta - dTheta / 2;
  
  c = cos(thetaMoy);
  s = sin(thetaMoy);

  /* jacobiennes */
  
  /* premiere ligne jxr */
  jxr[0] = 1.0;
  jxr[1] = 0;
  jxr[2] = -dS * s;
  /* deuxieme ligne jxr */
  jxr[3] = 0;
  jxr[4] = 1.0;
  jxr[5] = dS * c;
  /* troisieme ligne jxr */
  jxr[6] = 0;
  jxr[7] = 0;
  jxr[8] = 1.0;
  
  /* premiere ligne de jdep */
  jdep[0] = c/2. + dS*s/LaX;
  jdep[1] = c/2. - dS*s/LaX;
  /* deuxieme ligne de jdep */
  jdep[2] = s/2. - dS*c/LaX;
  jdep[3] = s/2. + dS*c/LaX;
  /* troisieme ligne de jdep */
  jdep[4] = -1/LaX;
  jdep[5] = 1/LaX;
}


/* variance sur le deplacement des roues; je suppose qu'il existe un
   facteur commun de proportion sur le deplacement; peut etre pas vrai;
   je pense qu'un veritable modele devrait contenir un terme constant
   plus un terme proportionel (quoi que si le robot ne bouge pas ... le
   terme constant devient plus que discutable !) 

   Il faudrait surtout un terme proportionnel a l'acceleration 
   (source de glissement)
*/

static void Calc_var_whee(double ds, double dth, double lax, 
			  double coeff, double *vdep)
{
  double dSR, dSL;

  dSR = ds + dth*lax/2;
  dSL = ds - dth*lax/2;
  vdep[0] = coeff * dSL * dSL;
  vdep[1] = 0;
  vdep[2] = coeff * dSR * dSR;
}



#define MATELT(mat,i,j,c)   ( *((mat) + (c) * (i) + (j)) )
#define INDEX_SMATELT(i,j)   \
((i)<(j) ? ((j)*((j)+1))/2 + (i) : ((i)*((i)+1))/2 +(j))
#define SMATELT(mat,i,j,n) *((mat) + INDEX_SMATELT((i),(j)))
#define MATSIZE(l,c)        ((l) * (c))
#define SMATSIZE(n,c)       (((n) * ((n) + 1)) / 2)

static int
dms_plus(const double *m1, const double *m2, double *m_res, int l, int c)
{
	int k;

	if (l != c)
	{
		fprintf (stderr, "dms : matrice symetrique non carree !!!\n");
		return -1;
	}

	for (k = 0; k < SMATSIZE(l, c); k++)
		*m_res++ = *m1++ + *m2++;
	return 0;
}

static int
dmfsf_jmult2 (const double *m1, const double *m2, double *m_res,
		int l1, int c1, int l2, int c2)
{

	double *temp, *res;
	double *aux = NULL;
	double *aux2 = NULL;
	int i,j,k,flag = 0;
	double val;

	if (l2 != c2 || c1 != c2)
	{
		fprintf(stderr, "%s : tailles incorrectes \n", __func__);
		return -1;
	}

	if (m_res == m1 || m_res == m2)
	{
		temp = malloc((l1*c2 + l1*(l1+1)/2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = temp + l1*c2;
		aux2 = res;
		flag = 1;
	}
	else
	{
		temp = malloc((l1 * c2) * sizeof(double));
		if (temp == NULL) {
			fprintf(stderr, "%s : Can't allocate memory for temp\n", __func__);
			return -1;
		}
		aux = temp;
		res = m_res;
	}


	for (i = 0; i < l1; i++)
		for (j = 0; j < c2; j++)
		{
			val = 0.;
			for (k = 0; k < c1; k++)
				val += MATELT (m1, i, k, c1) * SMATELT (m2, k, j, c2);
			*temp++ = val;
		}
	temp = aux;

	for (j = 0; j < l1; j++)
		for (i = 0; i <= j; i++)
		{
			val = 0.;
			for (k = 0; k < c2; k++)
				val += MATELT (temp, i, k, c2) * MATELT (m1, j, k, c1);
			*res++ = val;
		}

	if (flag)
	{
		res = aux2;
		for (i = 0; i < SMATSIZE (l1, l1); i++)
			*m_res++ = *res++;
	}

	free(temp);

	return 0;
}
